#include "sdk_common.h"
#include "emg_service_client_new.h"

#include <stdlib.h>
#include "ble.h"
#include "ble_db_discovery.h"
#include "ble_types.h"
#include "ble_srv_common.h"
#include "ble_gattc.h"
#include "app_error.h"
#define NRF_LOG_MODULE_NAME emg_service_client
#include "nrf_log.h"
NRF_LOG_MODULE_REGISTER();


void ble_emg_on_db_disc_evt(ble_emg_t * p_ble_emg, ble_db_discovery_evt_t * p_evt)
{
    ble_emg_evt_t emg_evt;
    memset(&emg_evt,0,sizeof(ble_emg_evt_t));

    ble_gatt_db_char_t * p_chars = p_evt->params.discovered_db.charateristics;

    // Check if the NUS was discovered.
    if (    (p_evt->evt_type == BLE_DB_DISCOVERY_COMPLETE)
        &&  (p_evt->params.discovered_db.srv_uuid.uuid == BLE_UUID_EMG_SERVICE_UUID)
        &&  (p_evt->params.discovered_db.srv_uuid.type == p_ble_emg->uuid_type))
    {
        for (uint32_t i = 0; i < p_evt->params.discovered_db.char_count; i++)
        {
            switch (p_chars[i].characteristic.uuid.uuid)
            {
         
                case BLE_UUID_EMG_CHAR_UUID:
                    emg_evt.handles.emg_handle = p_chars[i].characteristic.handle_value;
                    emg_evt.handles.emg_cccd_handle = p_chars[i].cccd_handle;
                    break;

                default:
                    break;
            }
        }
        if (p_ble_emg->evt_handler != NULL)
        {
            emg_evt.conn_handle = p_evt->conn_handle;
            emg_evt.evt_type    = BLE_EMG_SERVICE_CLIENT_EVT_DISCOVERY_COMPLETE;
            p_ble_emg->evt_handler(p_ble_emg, &emg_evt);
        }
    }
}

static void on_hvx(ble_emg_t * p_ble_emg, ble_evt_t const * p_ble_evt)
{
    // HVX can only occur from client sending.
    if (   (p_ble_emg->handles.emg_handle != BLE_GATT_HANDLE_INVALID)
        && (p_ble_evt->evt.gattc_evt.params.hvx.handle == p_ble_emg->handles.emg_handle)
        && (p_ble_emg->evt_handler != NULL))
    {
        ble_emg_evt_t ble_emg_evt;

        ble_emg_evt.evt_type = BLE_EMG_EVT_EMG_EVT;
        ble_emg_evt.p_data   = (uint8_t *)p_ble_evt->evt.gattc_evt.params.hvx.data;
        ble_emg_evt.data_len = p_ble_evt->evt.gattc_evt.params.hvx.len;

        p_ble_emg->evt_handler(p_ble_emg, &ble_emg_evt);
        //NRF_LOG_DEBUG("Client sending data.");
    }
}

uint32_t ble_emg_init(ble_emg_t * p_ble_emg, ble_emg_init_t * p_ble_emg_init)
{
    uint32_t      err_code;
    ble_uuid_t    emg_uuid;
    ble_uuid128_t emg_base_uuid = BLE_UUID_EMG_SERVICE_BASE_UUID;

    VERIFY_PARAM_NOT_NULL(p_ble_emg);
    VERIFY_PARAM_NOT_NULL(p_ble_emg_init);

    err_code = sd_ble_uuid_vs_add(&emg_base_uuid, &p_ble_emg->uuid_type);
    VERIFY_SUCCESS(err_code);

    emg_uuid.type = p_ble_emg->uuid_type;
    emg_uuid.uuid = BLE_UUID_EMG_SERVICE_UUID;

    p_ble_emg->conn_handle           = BLE_CONN_HANDLE_INVALID;
    p_ble_emg->evt_handler           = p_ble_emg_init->evt_handler;
    p_ble_emg->handles.emg_handle = BLE_GATT_HANDLE_INVALID;

    return ble_db_discovery_evt_register(&emg_uuid);
}

void ble_emg_on_ble_evt(ble_evt_t const * p_ble_evt, void * p_context)
{
    ble_emg_t * p_ble_emg = (ble_emg_t *)p_context;

    if ((p_ble_emg == NULL) || (p_ble_evt == NULL))
    {
        return;
    }

    if ( (p_ble_emg->conn_handle != BLE_CONN_HANDLE_INVALID)
       &&(p_ble_emg->conn_handle != p_ble_evt->evt.gap_evt.conn_handle)
       )
    {
        return;
    }

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GATTC_EVT_HVX:
            on_hvx(p_ble_emg, p_ble_evt);
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            if (p_ble_evt->evt.gap_evt.conn_handle == p_ble_emg->conn_handle
                    && p_ble_emg->evt_handler != NULL)
            {
                ble_emg_evt_t emg_evt;

                emg_evt.evt_type = BLE_EMG_EVT_DISCONNECTED;

                p_ble_emg->conn_handle = BLE_CONN_HANDLE_INVALID;
                p_ble_emg->evt_handler(p_ble_emg, &emg_evt);
            }
            break;

        default:
            // No implementation needed.
            break;
    }
}

static uint32_t cccd_configure(uint16_t conn_handle, uint16_t cccd_handle, bool enable)
{
    uint8_t buf[BLE_CCCD_VALUE_LEN];

    buf[0] = enable ? BLE_GATT_HVX_NOTIFICATION : 0;
    buf[1] = 0;

    ble_gattc_write_params_t const write_params =
    {
        .write_op = BLE_GATT_OP_WRITE_REQ,
        .flags    = BLE_GATT_EXEC_WRITE_FLAG_PREPARED_WRITE,
        .handle   = cccd_handle,
        .offset   = 0,
        .len      = sizeof(buf),
        .p_value  = buf
    };

    return sd_ble_gattc_write(conn_handle, &write_params);
}

uint32_t ble_emg_notif_enable(ble_emg_t * p_ble_emg)
{
    VERIFY_PARAM_NOT_NULL(p_ble_emg);

    if ( (p_ble_emg->conn_handle == BLE_CONN_HANDLE_INVALID)
       ||(p_ble_emg->handles.emg_cccd_handle == BLE_GATT_HANDLE_INVALID)
       )
    {
        return NRF_ERROR_INVALID_STATE;
    }
    return cccd_configure(p_ble_emg->conn_handle,p_ble_emg->handles.emg_cccd_handle, true);
}

uint32_t ble_emg_handles_assign(ble_emg_t                 * p_ble_emg,
                                  uint16_t                    conn_handle,
                                  ble_emg_handles_t const   * p_peer_handles)
{
    VERIFY_PARAM_NOT_NULL(p_ble_emg);

    p_ble_emg->conn_handle = conn_handle;
    if (p_peer_handles != NULL)
    {
        p_ble_emg->handles.emg_cccd_handle = p_peer_handles->emg_cccd_handle;
        p_ble_emg->handles.emg_handle      = p_peer_handles->emg_handle;
    }
    return NRF_SUCCESS;
}