#ifndef LED_SERVICE_CLIENT_H__
#define LED_SERVICE_CLIENT_H__

#include <stdint.h>
#include "ble.h"
#include "ble_db_discovery.h"
#include "nrf_sdh_ble.h"

#define BLE_EMG_BLE_OBSERVER_PRIO 2
#define BLE_EMG_DEF(_name)                                                                        \
static ble_emg_t _name;                                                                           \
NRF_SDH_BLE_OBSERVER(_name ## _obs,                                                                 \
                     BLE_EMG_BLE_OBSERVER_PRIO,                                                   \
                     ble_emg_on_ble_evt, &_name)


#define OPCODE_LENGTH 1
#define HANDLE_LENGTH 2

/**@brief   Maximum length of data (in bytes) that can be transmitted to the peer by the Nordic UART service module. */
#if defined(NRF_SDH_BLE_GATT_MAX_MTU_SIZE) && (NRF_SDH_BLE_GATT_MAX_MTU_SIZE != 0)
    #define BLE_EMG_MAX_DATA_LEN (NRF_SDH_BLE_GATT_MAX_MTU_SIZE - OPCODE_LENGTH - HANDLE_LENGTH)
#else
    #define BLE_NUS_MAX_DATA_LEN (BLE_GATT_MTU_SIZE_DEFAULT - OPCODE_LENGTH - HANDLE_LENGTH)
    #warning NRF_SDH_BLE_GATT_MAX_MTU_SIZE is not defined.
#endif

#define BLE_UUID_EMG_SERVICE_BASE_UUID              {{0x23, 0xD1, 0x13, 0xEF, 0x5F, 0x78, 0x23, 0x15, 0xDE, 0xEF, 0x12, 0x12, 0x00, 0x00, 0x00, 0x00}} // 128-bit base UUID
#define BLE_UUID_EMG_SERVICE_UUID                0xF00D // Just a random, but recognizable value
#define  BLE_UUID_EMG_CHAR_UUID          0xBEEF // Just a random, but recognizable value

//Event notifiying the discovery of EMG service at the peer
typedef enum
{
    BLE_EMG_SERVICE_CLIENT_EVT_DISCOVERY_COMPLETE = 1, 
    BLE_EMG_EVT_EMG_EVT,
    BLE_EMG_EVT_DISCONNECTED 
} ble_emg_evt_type_t;

//Handle of EMG characterisitc provided by soft device
/**@brief Handles on the connected peer device needed to interact with it. */
typedef struct
{
    uint16_t emg_handle;      /**< Handle of the EMG data. */
    uint16_t emg_cccd_handle; /**< Handle of the CCCD of the EMG characteristic as provided by a discovery. */
} ble_emg_handles_t;

typedef struct
{
    ble_emg_evt_type_t evt_type;
    uint16_t             conn_handle;
    uint16_t             max_data_len;
    uint8_t            * p_data;
    uint16_t             data_len;
    ble_emg_handles_t  handles;     /**< Handles on which the Nordic Uart service characteristics was discovered on the peer device. This will be filled if the evt_type is @ref BLE_NUS_C_EVT_DISCOVERY_COMPLETE.*/
} ble_emg_evt_t;

// Forward declaration of the ble_emg_t type.
typedef struct ble_emg_s ble_emg_t;

typedef void (* ble_emg_evt_handler_t)(ble_emg_t * p_ble_emg, ble_emg_evt_t const * p_evt);

/**@brief EMG Client structure. */
struct ble_emg_s
{
    uint8_t               uuid_type;      /**< UUID type. */
    uint16_t              conn_handle;    /**< Handle of the current connection. Set with @ref ble_nus_c_handles_assign when connected. */
    ble_emg_handles_t     handles;        /**< Handles on the connected peer device needed to interact with it. */
    ble_emg_evt_handler_t evt_handler;    /**< Application event handler to be called when there is an event related to the NUS. */
};

/**@brief EMG Client initialization structure. */
typedef struct
{
    ble_emg_evt_handler_t evt_handler;
} ble_emg_init_t;

//Function for the Initialize of EMG module
uint32_t ble_emg_init(ble_emg_t * p_ble_emg, ble_emg_init_t * p_ble_emg_init);

//Function for the handling of events from Database module
void ble_emg_on_db_disc_evt(ble_emg_t * p_ble_emg, ble_db_discovery_evt_t * p_evt);

 //Function for handling BLE events from Softdevice
void ble_emg_on_ble_evt(ble_evt_t const * p_ble_evt, void * p_context);

//Function for requesting the peer to start sending notification of EMG characteristic.
uint32_t ble_emg_notif_enable(ble_emg_t * p_ble_emg);

/**@brief Function for assigning handles to a this instance of nus_c.
 *
 * @details Call this function when a link has been established with a peer to
 *          associate this link to this instance of the module. This makes it
 *          possible to handle several link and associate each link to a particular
 *          instance of this module. The connection handle and attribute handles will be
 *          provided from the discovery event @ref BLE_NUS_C_EVT_DISCOVERY_COMPLETE.
 *
 * @param[in] p_ble_nus_c    Pointer to the NUS client structure instance to associate with these
 *                           handles.
 * @param[in] conn_handle    Connection handle to associated with the given NUS Instance.
 * @param[in] p_peer_handles Attribute handles on the NUS server that you want this NUS client to
 *                           interact with.
 *
 * @retval    NRF_SUCCESS    If the operation was successful.
 * @retval    NRF_ERROR_NULL If a p_nus was a NULL pointer.
 */
uint32_t ble_emg_handles_assign(ble_emg_t *               p_ble_emg,
                                  uint16_t                    conn_handle,
                                  ble_emg_handles_t const * p_peer_handles);

#endif